SystemOrganization addCategory: #'SqueakNxt-Compiler-Programs'!
SystemOrganization addCategory: #'SqueakNxt-Compiler-Translation'!
SystemOrganization addCategory: #'SqueakNxt-Compiler-Api'!

Object subclass: #Nbc
	instanceVariableNames: 'parameters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Api'!

!Nbc commentStamp: 'Richo 12/25/2010 01:12' prior: 0!
A NbcExe is a very simple API for accessing the nbc compiler using FFI on Windows

Examples: 
1) The simplest example would be compiling into a file called "output" and reporting errors to a file called "errors":

	NbcExe new
		fileName: 'example_program.nbc';
		outFile: 'output';
		errorsFile: 'errors';
		execute.
		
2) To download a program into the Lego Nxt using the usb:

	NbcExe new
		fileName: 'example_program.nbc';
		portName: 'usb';
		downloadProgram;
		execute.
		
3) To use the bluetooth on COM6 instead:
	
	NbcExe new
		fileName: 'example_program.nbc';
		useBluetooth;
		portName: 'COM6';
		downloadProgram;
		execute.

Instance Variables
	parameters:		<WriteStream>

parameters
	- Used to build the parameters string to be sent to the nbc program. See "nbc -help" for more information.
!

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:45'!
beQuiet
	parameters space; nextPutAll: '-q'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:45'!
decompileProgram
	parameters space; nextPutAll: '-x'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:45'!
defineMacro: aByteString 
	parameters space; nextPutAll: '-D='; nextPutAll: aByteString ! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:45'!
dontCompile
	parameters space; nextPutAll: '-b'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:45'!
dontIncludeSystemFile
	parameters space; nextPutAll: '-n'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:46'!
downloadProgram
	parameters space; nextPutAll: '-d'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/28/2010 18:25'!
enhancedFirmware
	parameters space; nextPutAll: '-EF'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:46'!
errorsFile: aByteString 
	parameters space; nextPutAll: '-E='; nextPutAll: aByteString ! !

!Nbc methodsFor: 'executing' stamp: 'Richo 12/29/2010 00:38'!
execute
	"Executes the command using FFI. Windows only for now"
	| result shell |
	shell := Smalltalk at: #Win32Shell ifAbsent: [^self error: 'FFI doesn''t seem to be installed'].
	(result := (shell new 
		shellExecute:  nil
		lpOperation:  'open'
		lpFile:  'nbc.exe'
		lpParameters:  parameters contents
		lpDirectory:  nil
		nShowCmd: 0) <= 32) ifTrue: [self error: 'system error, code:', result]! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:46'!
fileName: aByteString 
	parameters space; nextPutAll: aByteString! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:47'!
includePath: aByteString 
	parameters space; nextPutAll: '-I='; nextPutAll: aByteString ! !

!Nbc methodsFor: 'initialize' stamp: 'Richo 12/25/2010 00:44'!
initialize
	super initialize.
	parameters := WriteStream on: ''! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:47'!
outFile: aByteString 
	parameters space; nextPutAll: '-O='; nextPutAll: aByteString ! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:47'!
portName: aByteString 
	parameters space; nextPutAll: '-S='; nextPutAll: aByteString ! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/25/2010 00:48'!
useBluetooth
	parameters space; nextPutAll: '-BT'! !

!Nbc methodsFor: 'accessing' stamp: 'Richo 12/28/2010 18:38'!
warningsOn
	parameters space; nextPutAll: '-w'! !

Object subclass: #NbcCodeBlock
	instanceVariableNames: 'statements name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcCodeBlock class methodsFor: 'instance creation' stamp: 'Richo 12/28/2010 17:46'!
name: aString statements: statements
	^ (self basicNew setName: aString statements: statements) initialize! !

!NbcCodeBlock methodsFor: 'printing' stamp: 'Richo 12/28/2010 17:48'!
printOn: aStream
aStream nextPutAll: name; cr.
statements do: [:each | aStream tab.
	each printOn: aStream.
	aStream cr].! !

!NbcCodeBlock methodsFor: 'private' stamp: 'Richo 12/28/2010 17:47'!
setName: aString statements: anOrderedCollection 
	name := aString.
	statements := anOrderedCollection ! !

NbcCodeBlock subclass: #NbcSubroutine
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcSubroutine methodsFor: 'printing' stamp: 'Richo 12/28/2010 17:49'!
printOn: aStream
aStream nextPutAll: 'subroutine'; space.
super printOn: aStream.
aStream nextPutAll: 'return'; cr; nextPutAll: 'ends'! !

NbcCodeBlock subclass: #NbcThread
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcThread methodsFor: 'printing' stamp: 'Richo 12/28/2010 17:46'!
printOn: aStream
aStream nextPutAll: 'thread'; space.
super printOn: aStream.
aStream nextPutAll: 'endt'! !

Object subclass: #NbcCompiler
	instanceVariableNames: 'programName source'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Api'!

!NbcCompiler commentStamp: 'Richo 12/27/2010 18:33' prior: 0!
A NbcCompiler is responsible for verifying/compiling programs as well as downloading them to the LegoNxt.

===============================
	"Code example"
	
	| program |
	
	"This nbc program simply plays five tones on the nxt"
	program := '
		thread main
			PlayTone(TONE_C5, 500)
			wait 500
			PlayTone(TONE_D5, 500)
			wait 500
			PlayTone(TONE_E5, 500)
			wait 500
			PlayTone(TONE_D5, 500)
			wait 500
			PlayTone(TONE_C5, 1000)
			wait 1000
		endt'.

	"The code below will verify the program and return a collection of errors (if any)"
	NbcCompiler new
		source: program;
		verify.
	
	"The code below will download the program to the nxt using the usb cable. the resulting program will be called 'PlayingMusic'"
	NbcCompiler new
		programName: 'PlayingMusic';
		source: program;
		downloadUsingUsb.
===============================

Instance Variables
	programName:		<String>
	source:		<String>

programName
	- A name for the program. If not specified the program name will be the NbcCompiler instance's hash (not a nice name for a program, I must say).

source
	- The nbc source code of the program.
!

!NbcCompiler methodsFor: 'compiling' stamp: 'Richo 12/27/2010 22:10'!
compileInFileNamed: outFile
	"Compiles the program into a file"
	^ self execute: (Nbc new outFile: outFile)! !

!NbcCompiler methodsFor: 'compiling' stamp: 'Richo 12/27/2010 18:10'!
downloadUsingBluetooth: portName 
	"Compiles the program and downloads it using the bluetooth on the specified port name (COMn). Returns a collection of NbcErrors"
	^ self execute: (Nbc new portName: portName; useBluetooth; downloadProgram)! !

!NbcCompiler methodsFor: 'compiling' stamp: 'Richo 12/27/2010 18:09'!
downloadUsingUsb
	"Compiles the program and downloads it using the usb cable. Returns a collection of NbcErrors"
	^ self execute: (Nbc new portName: 'usb'; downloadProgram)! !

!NbcCompiler methodsFor: 'private' stamp: 'Richo 12/27/2010 18:14'!
errorsFileName
	"Temporary file name to store errors during compilation"
	^ self programName , '.err'! !

!NbcCompiler methodsFor: 'private' stamp: 'Richo 12/28/2010 18:38'!
execute: nbc 
	"Create the temporary files to store the source code and the errors.
	Execute nbc.
	Wait until nbc finishes and parse errors.
	Delete temporary files.
	Return errors"
	source isNil ifTrue: [^ self error: 'Source cannot be nil'].
	FileStream forceNewFileNamed: self programFileName do: [:file | file nextPutAll: source]; 
				forceNewFileNamed: self errorsFileName do: [:file | "Do nothing. Nbc will write this file"].
	[nbc fileName: self programFileName;
		 errorsFile: self errorsFileName;
		 execute.
	^ self parseErrors]
		ensure: [FileDirectory default deleteFileNamed: self programFileName;
				 deleteFileNamed: self errorsFileName]! !

!NbcCompiler methodsFor: 'private' stamp: 'Richo 12/27/2010 18:12'!
parse: errorsFile
"Returns a collection with all the errors founded by the nbc compiler. The errorsFile format is similar to this:

# Error: Unknown or invalid statement
File ""D:\RICHI\Squeak images\Nxt\simpleprogram.nbc"" ; line 7
#   result sbyt
#----------------------------------------------------------
1 error during compilation
"
| errors |
errors := OrderedCollection new.
[errorsFile peek = $#] 
	whileTrue: [ | errorDescription errorPosition errorSource |
		errorDescription := errorsFile nextLine allButFirst: 2.
		errorPosition := (errorsFile nextLine subStrings: {$;}) last withoutLeadingBlanks.
		errorSource := errorsFile nextLine allButFirst: 4.
		errorsFile nextLine.
		errors add: (NbcError description: errorDescription position: errorPosition source: errorSource).
		].
^errors asArray! !

!NbcCompiler methodsFor: 'private' stamp: 'Richo 12/27/2010 22:05'!
parseErrors
	"Wait until the errors file is fully written by nbc, then read it and return the errors.
	IMPORTANT: Be aware that this code relies on the fact that FileStream>>#fileNamed: returns nil when the file is being modified by another process. If this somehow changes in the future, this code will inevitably break."
	| errorsFile |
	[500 milliSeconds asDelay wait.
	(errorsFile := FileStream fileNamed: self errorsFileName) isNil] whileTrue.
	[^ self parse: errorsFile]
		ensure: [errorsFile close].
! !

!NbcCompiler methodsFor: 'accessing' stamp: 'Richo 12/29/2010 00:48'!
program: aNbcProgram
	source := aNbcProgram asString! !

!NbcCompiler methodsFor: 'private' stamp: 'Richo 12/27/2010 18:14'!
programFileName
	"Temporary file name to store the source code"
	^ self programName , '.nbc'! !

!NbcCompiler methodsFor: 'accessing' stamp: 'Richo 12/27/2010 22:10'!
programName
	"If programName is not specified, the default name will be this instance's hash"
	^ programName ifNil: [self hash asString]! !

!NbcCompiler methodsFor: 'accessing' stamp: 'Richo 12/27/2010 16:47'!
programName: aString
	programName := aString! !

!NbcCompiler methodsFor: 'accessing' stamp: 'Richo 12/27/2010 17:07'!
source
	^ source! !

!NbcCompiler methodsFor: 'accessing' stamp: 'Richo 12/27/2010 16:48'!
source: aString
	source := aString.! !

!NbcCompiler methodsFor: 'compiling' stamp: 'Richo 12/27/2010 22:10'!
verify
	"Compiles the program and returns a collection of NbcErrors. If the collection is empty then the program is ok"
	^ self execute: Nbc new! !

Object subclass: #NbcError
	instanceVariableNames: 'description position source'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Api'!

!NbcError class methodsFor: 'instance creation' stamp: 'Richo 12/25/2010 16:02'!
description: description position: position source: source
^ (self basicNew setDescription: description position: position source: source) initialize ! !

!NbcError methodsFor: 'accessing' stamp: 'Richo 12/27/2010 16:01'!
description
	^ description ! !

!NbcError methodsFor: 'accessing' stamp: 'Richo 12/27/2010 16:01'!
position
	^ position ! !

!NbcError methodsFor: 'printing' stamp: 'Richo 12/27/2010 22:08'!
printOn: aStream
aStream nextPutAll: position asString;
		 nextPut: $:;
		 space;
		 nextPutAll: description asString 
! !

!NbcError methodsFor: 'private' stamp: 'Richo 12/25/2010 16:02'!
setDescription: aByteString position: aByteString2 source: aByteString3 
	description := aByteString.
	position := aByteString2.
	source := aByteString3! !

!NbcError methodsFor: 'accessing' stamp: 'Richo 12/27/2010 16:01'!
source
	^ source ! !

Object subclass: #NbcProgram
	instanceVariableNames: 'variables codeBlocks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcProgram class methodsFor: 'instance creation' stamp: 'Richo 12/28/2010 17:02'!
variables: aSet codeBlocks: anArray 
	^ (self basicNew setVariables: aSet codeBlocks: anArray ) initialize ! !

!NbcProgram methodsFor: 'printing' stamp: 'Richo 12/28/2010 17:48'!
printOn: aStream
aStream nextPutAll: 'dseg segment'; cr.
variables do: [:each|
	each = 'self' ifFalse: [aStream tab; nextPutAll: each; space; nextPutAll: 'slong'; cr].
	].
aStream nextPutAll: 'dseg ends'; cr; cr.
codeBlocks do: [:each | each printOn: aStream. aStream cr; cr]! !

!NbcProgram methodsFor: 'private' stamp: 'Richo 12/28/2010 17:02'!
setVariables: aSet codeBlocks: anArray 
	variables := aSet.
	codeBlocks := anArray.! !

Object subclass: #NbcStatement
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

NbcStatement subclass: #NbcAssignmentStatement
	instanceVariableNames: 'variable value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcAssignmentStatement class methodsFor: 'instance creation' stamp: 'Richo 12/28/2010 15:32'!
variable: aByteString value: aByteString2 
^( self basicNew setVariable: aByteString value: aByteString2) initialize ! !

!NbcAssignmentStatement methodsFor: 'printing' stamp: 'Richo 12/28/2010 15:34'!
printOn: aStream
aStream nextPutAll: 'mov'; space; nextPutAll: variable; nextPut: $,; space; nextPutAll: value! !

!NbcAssignmentStatement methodsFor: 'private' stamp: 'Richo 12/28/2010 15:32'!
setVariable: aByteString value: aByteString2 
	variable := aByteString.
	value := aByteString2 ! !

NbcStatement subclass: #NbcMacroFunctionCall
	instanceVariableNames: 'arguments rule'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcMacroFunctionCall class methodsFor: 'instance creation' stamp: 'Richo 12/28/2010 15:37'!
rule: aByteSymbol arguments: anArray 
	^ (self basicNew setRule: aByteSymbol arguments: anArray ) initialize ! !

!NbcMacroFunctionCall methodsFor: 'printing' stamp: 'Richo 12/28/2010 19:13'!
printOn: aStream
| tempRule |
tempRule := rule.
arguments doWithIndex: [:arg :i | 
tempRule := tempRule copyReplaceAll: 'arg', (i - 1) asString with: arg].
aStream nextPutAll: tempRule.! !

!NbcMacroFunctionCall methodsFor: 'private' stamp: 'Richo 12/28/2010 15:38'!
setRule: aByteSymbol arguments: anArray 
	rule := aByteSymbol.
	arguments := anArray ! !

NbcStatement subclass: #NbcMathStatement
	instanceVariableNames: 'variable arguments name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcMathStatement class methodsFor: 'instance creation' stamp: 'Richo 12/28/2010 15:35'!
name: aByteSymbol variable: aByteString arguments: anArray 
	^ (self basicNew setName: aByteSymbol variable: aByteString arguments: anArray ) initialize! !

!NbcMathStatement methodsFor: 'printing' stamp: 'Richo 12/28/2010 15:36'!
printOn: aStream
aStream nextPutAll: name; space;
		nextPutAll: variable; nextPut: $,; space.
arguments do: [:arg | aStream nextPutAll: arg] separatedBy: [aStream nextPut: $,; space]! !

!NbcMathStatement methodsFor: 'private' stamp: 'Richo 12/28/2010 15:36'!
setName: aByteSymbol variable: aByteString arguments: anArray 
	name := aByteSymbol.
	variable := aByteString.
	arguments := anArray ! !

NbcStatement subclass: #NbcTimingStatement
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcTimingStatement class methodsFor: 'instance creation' stamp: 'Richo 12/28/2010 16:52'!
value: aByteString 
	^ (self basicNew setValue: aByteString) initialize ! !

!NbcTimingStatement methodsFor: 'printing' stamp: 'Richo 12/28/2010 16:53'!
printOn: aStream
aStream nextPutAll: 'waitv'; space; nextPutAll: value! !

!NbcTimingStatement methodsFor: 'private' stamp: 'Richo 12/28/2010 16:52'!
setValue: aByteString 
	value := aByteString ! !

Object subclass: #NxtProgram
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Programs'!

!NxtProgram class methodsFor: 'translating' stamp: 'Richo 12/29/2010 00:39'!
asNbcProgram
^ NbcTranslator new translate: self! !

!NxtProgram methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 17:25'!
displayNumber: var x: x y: y! !

!NxtProgram methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 17:25'!
onFwd: outputConstant speed: speed! !

!NxtProgram methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 17:25'!
onRev: outputConstant speed: speed! !

!NxtProgram methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 17:25'!
wait: milliseconds! !

NxtProgram subclass: #NxtProgramTest1
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Programs'!

!NxtProgramTest1 methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 17:42'!
anotherMethod
	self displayNumber: 100 x: 10 y: 10! !

!NxtProgramTest1 methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 17:41'!
main
	self onFwd: 0 speed: 100;
		wait: 1000.! !

NxtProgram subclass: #NxtProgramTest2
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Programs'!

!NxtProgramTest2 methodsFor: 'as yet unclassified' stamp: 'Richo 12/28/2010 18:46'!
main
	| var |
	var := 10 - 20 * (30 * 40).
	self displayNumber: var x: 10 y: 8;
		wait: var;
		onFwd: 0 speed: 100! !

RBProgramNodeVisitor subclass: #NbcTranslator
	instanceVariableNames: 'statements variables codeBlocks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SqueakNxt-Compiler-Translation'!

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 15:24'!
acceptAssignmentNode: anAssignmentNode
| value variable |
variable := self visitNode: anAssignmentNode variable.
value := self visitNode: anAssignmentNode value.
statements add: (NbcAssignmentStatement variable: variable value: value).
^ variable! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 15:06'!
acceptLiteralNode: aLiteralNode
	^aLiteralNode value asString! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 15:22'!
acceptMessageNode: aMessageNode
^ self createStatementFrom: aMessageNode! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 16:30'!
acceptVariableNode: aVariableNode
	^ variables add: aVariableNode name
	! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 17:28'!
createMacroFunctionFrom: aMessageNode 
	| arguments receiver selector |
	selector := aMessageNode selector.
	receiver := self visitNode: aMessageNode receiver.
	arguments := aMessageNode arguments collect: [:each | self visitNode: each].
	statements add: (NbcMacroFunctionCall rule: (self macroFunctions at: selector) arguments: {receiver} , arguments)! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 15:35'!
createMathStatementFrom: aMessageNode 
| argument receiver selector variable |
selector := aMessageNode selector.
variable := self newVariable.
receiver := self visitNode: aMessageNode receiver.
argument := self visitNode: (aMessageNode arguments first).
statements add: (NbcMathStatement name: (self mathStatements at: selector) variable: variable arguments: {receiver. argument}). 
^ variable.
! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 16:50'!
createStatementFrom: aMessageNode 
(self macroFunctions includesKey: aMessageNode selector)
	ifTrue: [^ self createMacroFunctionFrom: aMessageNode].
(self mathStatements includesKey: aMessageNode selector)
	ifTrue: [^ self createMathStatementFrom: aMessageNode].
(self timingStatements includesKey: aMessageNode selector)
	ifTrue: [^ self createTimingStatementFrom: aMessageNode].
self error: 'What the fuck?'! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 16:51'!
createTimingStatementFrom: aMessageNode 
	statements add: (NbcTimingStatement value: (self visitNode: aMessageNode arguments first)).! !

!NbcTranslator methodsFor: 'accessing' stamp: 'Richo 12/28/2010 17:24'!
macroFunctions
	^ {
	#displayNumber:x:y: -> 'NumOut(arg2, arg3, arg1)'.
	#onFwd:speed: -> 'OnFwd(arg1, arg2)'
	} as: Dictionary! !

!NbcTranslator methodsFor: 'accessing' stamp: 'Richo 12/28/2010 17:30'!
mathStatements
^ {
	#+ -> 'add'.
	#- -> 'sub'.
	#* -> 'mul'
} as: Dictionary ! !

!NbcTranslator methodsFor: 'visiting' stamp: 'Richo 12/28/2010 16:47'!
newVariable
| auxVarName |
auxVarName := 't', self hash asString.
^variables add: (auxVarName , (variables select: [:each | auxVarName, '*' match: each]) size asString).! !

!NbcTranslator methodsFor: 'accessing' stamp: 'Richo 12/28/2010 16:50'!
timingStatements
^ {
	#wait: -> 'waitv'
} as: Dictionary ! !

!NbcTranslator methodsFor: 'translating' stamp: 'Richo 12/29/2010 00:39'!
translate: aClass 
	variables := Set new.
	codeBlocks := aClass selectors collect: [:each| 
			self translateMethod: (aClass sourceCodeAt: each) named: each asString.
		].
	^ NbcProgram variables: variables codeBlocks: codeBlocks! !

!NbcTranslator methodsFor: 'translating' stamp: 'Richo 12/28/2010 17:47'!
translateMethod: methodSource named: methodName
statements := OrderedCollection new.
self visitNode: (RBParser parseMethod: methodSource).
^((methodName = 'main') ifTrue: [NbcThread] ifFalse: [NbcSubroutine])
	name: methodName statements: statements! !
